/*******************************************************************************
 * JMMC project ( http://www.jmmc.fr ) - Copyright (C) CNRS.
 ******************************************************************************/

/***************************************************************************/

local package_amdlibPlugin;
/* DOCUMENT package_amdlibPlugin -- amdlibPlugin.i
   Low level amdlib functions
*/

// Wrapper to C sub-routines 
#include "amdlibWrapper.i"

/*
* UV Coordinates
*/
struct amdlibUV_COORD
{
    double u;
    double v;
};

/********************* AMBER Data Reduction routines *********************
*
* A complete amdlib interface has been generated automatically from the AMBER
* Data Reduction library, called amdlib, header file and is named
* amdlibWrapper.i 
* Here is the routines which have to be used to call the amdlib interface 
*/

/************************************************************************/

func amdlibPlugin(void)
/* DOCUMENT amdlibPlugin
    
  DESCRIPTION
    amdlib plugin low level routines

  VERSION
    $Revision: 13425 $
    
  FUNCTIONS
    - amdlibComputeVisibilities    : The amdlibComputeVisibilities()
                                     compute the raw visibilities and pistons.
    - amdlibGetOiArrayFromRawData  : Retrieve OI array structure from
                                     the specified raw data.
    - amdlibGetOiTargetFromRawData : Retrieve OI target structure from
                                     the specified raw data.
    - amdlibGetDarkData            : Gets the data stored in dark data structure
    - amdlibGetWaveDataFromP2vm    : Retrieve wave data structure from
                                     the specified p2vm.
    - amdlibPlugin                 : 
    - amdlibSetBadPixelMap         : Set all values to the bad pixel map
                                     to the given value.
    - amdlibSetFlatFieldMap        : Set all values to the flat field
                                     map to the given value.
    - amdlibSetDarkData            : replaces the data stored in dark data 
                                     structure with the one placed
                                     into the arrays of an amdlibREGION data
                                     structure.
    - amdlibSetRawData             : replaces the data stored in raw data
                                     structure with the one placed into the
                                     arrays of an amdlibREGION data structure.
    - amdlibSetScienceData         : 
    - amdlibVersion                : This function prints or gets the
                                     current version of amdlib.
    - amdlibWriteOiFile            : Writes OI files from the AMBER data
                                     contained in the amdlib structures

  SEE ALSO
    yocoFile, amdlib
*/
{
    version = strpart(strtok("$Revision: 13425 $",":")(2),2:-2);
    if (am_subroutine())
    {
        help, amdlibPlugin;
    }   
    return version;
}


/************************************************************************/

func amdlibVersion(&v)
/* DOCUMENT amdlibVersion(&v)
  
  DESCRIPTION 
    This function prints or gets the current version of amdlib.
  
  PARAMETERS
    - v : returned version; if not specified print out version

  SEE ALSO
*/
{
    verstr = array(char, amdlibERROR_LENGTH);
    __amdlibGetVersion, &verstr;
    verStr = string(&verstr);

    if (is_void(v))
    {
        write, "\"", verStr, "\"";
    }
    else
    {
        v = verStr;
    }
}

/************************************************************************/

func _amdlibLoadRawData(filename, &rawData, &errMsg)
/* DOCUMENT _amdlibLoadRawData(filename, &rawData, &errMsg)

  DESCRIPTION
    The raw data contains data from the detector with the amdlibRAW_DATA
    structure. This routine allows to load such data.

  PARAMETERS
    - filename : name of file to load
    - rawData  : structure where read raw data will be stored
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS 
    Structure in which raw data is stored has been dynamically allocated and
    must be released using _amdlibReleaseRawData() when not longer used
    or before futur use, otherwise this will cause unpredictable behaviour.

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    ...
    > _amdlibReleaseRawData, &rawData;

  SEE ALSO
    _amdlibLoadFlatFieldMap, _amdlibLoadBadPixelMap, 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    rawData = amdlibRAW_DATA();

    status = __amdlibLoadRawData(filename, &rawData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibLoadRawDataHdr(filename, &rawData, &errMsg)
/* DOCUMENT _amdlibLoadRawDataHdr(filename, &rawData, &errMsg)

  DESCRIPTION
    The raw data contains data from the detector with the amdlibRAW_DATA 
    structure. This routine allows to load data description information
    without loading data themselves. This is useful to handle large file, to
    know the number of frames contained in data, before loading it by packets. 

  PARAMETERS
    - filename : name of file to load
    - rawData  : structure where read raw data will be stored
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    Get the number of frames in file.fits

    > _amdlibLoadRawDataHdr("file.fits", rawData, errMsg);
    > write, "Number of frames = ", rawData.nbFrames
    250

  SEE ALSO
    _amdlibLoadRawData, _amdlibLoadRawFrames
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    rawData = amdlibRAW_DATA();

    status = __amdlibLoadRawDataHdr(filename, &rawData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibLoadRawFrames(filename, &rawData, firstFrame, nbFrames, &errMsg)
/* DOCUMENT _amdlibLoadRawFrames(filename, &rawData, firstFrame, nbFrames, &errMsg)

  DESCRIPTION
    The raw data contains data from the detector with the amdlibRAW_DATA
    structure. This routine allows to load the <nbFrames> frames from
    <firstFrame>th.

  PARAMETERS
    - filename   : name of file to load
    - rawData    : structure where read raw data will be stored
    - firstFrame : first frame to load
    - nbFrames   : number of frames to load
    - errMsg     : if function fails, error message describing failure
  
  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which raw data is stored has been dynamically allocated and
    must be released using _amdlibReleaseRawData() when not longer used
    or before futur use, otherwise this will cause unpredictable behaviour.
   
  EXAMPLES
    Load the 2nd and 3rd frames from file.fits

    > _amdlibLoadRawFrames("file.fits", rawData, 2, 2, errMsg);
    ...
    > _amdlibReleaseRawData, &rawData;

  SEE ALSO
    _amdlibLoadRawData, _amdlibLoadRawDataHdr
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    rawData = amdlibRAW_DATA();

    status = __amdlibLoadRawFrames(filename, &rawData, 
                                   int(firstFrame), int(nbFrames), &errMsg);
    errMsg = string(&errMsg);

    return status;
} 

/************************************************************************/

func _amdlibReleaseRawData(&rawData)
/* DOCUMENT _amdlibReleaseRawData(&rawData)

  DESCRIPTION
    Release memory allocated by _amdlibLoadRawData or _amdlibLoadRawFrames, to
    store raw data.

  PARAMETERS
    - rawData : structure to be released
  
  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    ...
    > _amdlibReleaseRawData, &rawData;

  SEE ALSO
    _amdlibLoadRawData, _amdlibLoadRawDataHdr
*/
{
    status = __amdlibReleaseRawData(rawData);
} 

/************************************************************************/

func _amdlibCalibrateRawData(&darkData, &rawData, &errMsg)
/* DOCUMENT _amdlibCalibrateRawData(&darkData, &rawData, &errMsg)

  DESCRIPTION
    The _amdlibCalibrateRawData() function applies cosmetic corrections on raw
    data. It uses the bad pixel and the flat-field maps respectively loaded
    with _amdlibLoadBadPixelMap() and _amdlibLoadFlatFieldMap().

  PARAMETERS
    - darkData  : structure where dark data is stored
    - rawData   : structure where raw data is stored
    - errMsg    : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadBadPixelMap("amdmsBadPixelMap.fits", errMsg);
    > _amdlibLoadFlatFieldMap("amdmsFlatFieldMap.fits", errMsg);
    > _amdlibLoadRawData("dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);
    ...
    > _amdlibReleaseDarkData, &darkData;
    > _amdlibReleaseRawData, &rawData;

  SEE ALSO
    _amdlibLoadFlatFieldMap, _amdlibLoadBadPixelMap, 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCalibrateRawData(darkData, rawData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibReleaseDarkData(&darkData)
/* DOCUMENT _amdlibReleaseDarkData(&darkData)

  DESCRIPTION
    Release memory allocated by _amdlibLoadRawData or _amdlibLoadRawFrames, to
    store raw data.

  PARAMETERS
    - darkData : structure where dark is stored
  
  EXAMPLES
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    ...
    > _amdlibReleaseDarkData, &darkData;

  SEE ALSO
    _amdlibLoadRawData, _amdlibLoadRawDataHdr
*/
{
    status = __amdlibReleaseDarkData(darkData) 
} 

/************************************************************************/

func _amdlibGetRawData(&rawData)
/* DOCUMENT _amdlibGetRawData(&rawData)

  DESCRIPTION
    This function gets the data stored in raw data structure and places them
    into the arrays of an amdlibREGION data structure which is returned by the
    function. 

  PARAMETERS
    - rawData : structure where raw data is stored

  RETURN VALUES
    A amdlibREGION structure with region data stored in

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg)
    > r = _amdlibGetRawData(&rawData);
    > _amdlibReleaseRawData, &rawData
    > pli, (*r(4).data)(, , 1)

  SEE ALSO
    _amdlibLoadRawData, amdlibSetRawData
*/
{
    /* First get the dimensions of regions */
    regionTab = array(amdlibREGION, rawData->nbRows, rawData->nbCols);
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyRegionInfo(rawData->region, &regionTab, 
                                    rawData->nbRegions, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy information region", errMsg, 3;
    }

    /* Then allocate memory and get the data*/
    for (i = 1; i <= rawData->nbRows; i++)
    {
        for (j = 1; j <= rawData->nbCols; j++)
        {
            regionTab(i, j).data = &array(double, 
                                          regionTab(i, j).dimAxis(1), 
                                          regionTab(i, j).dimAxis(2), 
                                          regionTab(i, j).dimAxis(3));
        }
    }

    status = __amdlibCopyRegionData(rawData->region, &regionTab, 
                                    rawData->nbRegions, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy data region", errMsg, 3;
    }
    return regionTab;
}

/************************************************************************/

func amdlibSetRawData(regionTab, &rawData)
/* DOCUMENT amdlibSetRawData(regionTab, &rawData)

  DESCRIPTION
    This function replaces the data stored in raw data structure with the one
    placed into the arrays of an amdlibREGION data structure.

  PARAMETERS
    - regionTab : region data structure containing new data values
    - rawData   : structure where raw data has to be stored. It is assumed
                  this structure has been previously allocated.

  RETURN VALUES

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg)
    > r=_amdlibGetRawData(&rawData);
    ...
    > amdlibSetRawData(r, &rawData);
    ...
    > _amdlibReleaseRawData, &rawData

  SEE ALSO
    _amdlibGetRawData
*/
{
    /* Copy region data into the raw data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyRegionData(&regionTab, rawData->region, 
                                    rawData->nbRegions, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy data region", errMsg, 3;
    }

    return amdlibSUCCESS;
}

/************************************************************************/

func _amdlibLoadBadPixelMap(filename, &errMsg)
/* DOCUMENT _amdlibLoadBadPixelMap(filename, &errMsg)

  DESCRIPTION
    The bad pixel map contains for each pixel a flag (1 = good pixel, 0 = bad
    pixel). All pixel marked as bad has to be ignored but the values for that
    pixels can be interpolated from the values of the surrounding pixels. Bad
    pixel map can be loaded from fits files.

  PARAMETERS
    - filename : name of file to load
    - errMsg   : if function fails, error message describing failure 

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadBadPixelMap("BadPixelMap.fits", errMsg);

  SEE ALSO
    amdlibSetBadPixelMap, _amdlibLoadFlatFieldMap 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);

    status = __amdlibLoadBadPixelMap(filename, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func amdlibSetBadPixelMap(value)
/* DOCUMENT amdlibSetBadPixelMap(value)

  DESCRIPTION
    Set all values to the bad pixel map to the given value.

  PARAMETERS
    - value : value for all pixel of the map; could be amdlibGOOD_PIXEL_FLAG
              or amdlibBAD_PIXEL_FLAG

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > amdlibSetBadPixelMap(amdlibGOOD_PIXEL_FLAG);
  
  SEE ALSO
    _amdlibLoadBadPixelMap 
*/
{
    status = __amdlibSetBadPixelMap(value);

    return status;
} 

/************************************************************************/

func _amdlibGenerateDarkData(&rawData, &darkData, &errMsg)
/* DOCUMENT _amdlibGenerateDarkData(&rawData, &darkData, &errMsg)

  DESCRIPTION
    The pixel bias map is needed in order to compensate for detector effects.
    This function computed the pixel bias from the given raw data.

  PARAMETERS
    - rawData   : structure where raw data is stored
    - darkData  : structure where dark will be stored
    - errMsg    : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which pixel bias data is stored has been dynamically
    allocated and must be released using _amdlibReleaseDarkData() when not
    longer used or before futur use, otherwise this will cause unpredictable
    behaviour.

  EXAMPLES
    > _amdlibLoadRawData("dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibReleaseDarkData, &darkData;
  
  SEE ALSO
    amdlibLoadRawData, _amdlibReleaseDarkData
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    darkData = amdlibDARK_DATA();

    status = __amdlibGenerateDarkData(&rawData, &darkData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func amdlibGetDarkData(&darkData)
/* DOCUMENT amdlibGetDarkData(&darkData)

  DESCRIPTION
    This function gets the data stored in pixel bias data structure and place
    them into the arrays of an amdlibREGION data structure which is returned
    by the function. 

  PARAMETERS
    - darkData : structure where i.e dark is stored

  RETURN VALUES
    A amdlibREGION structure with region data stored in

  EXAMPLES
    > _amdlibLoadRawData("dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > p=amdlibGetDarkData(&darkData);
    > _amdlibReleaseDarkData, &darkData;
    > pli, (*p(4).data)(, , 1)

  SEE ALSO
    amdlibLoadRawData, amdlibSetDarkData
*/
{
    /* First get the dimensions of regions */
    regionTab = array(amdlibREGION, darkData->nbRows, darkData->nbCols);
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyRegionInfo(darkData->region, &regionTab, 
                                    darkData->nbRegions, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy information region", errMsg, 3;
        return 0;
    }

    /* Then allocate memory and get the data*/
    for (i = 1; i <= darkData->nbRows; i++)
    {
        for (j = 1; j <= darkData->nbCols; j++)
        {
            idx = (i - 1) * darkData->nbCols + j;
            regionTab(i, j).data = &array(double, 
                                          regionTab(idx).dimAxis(1), 
                                          regionTab(idx).dimAxis(2), 
                                          regionTab(idx).dimAxis(3));
        }
    }  

    status = __amdlibCopyRegionData(darkData->region, &regionTab, 
                                    darkData->nbRegions, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy data region", errMsg, 3;
    }

    return regionTab;
}

/************************************************************************/

func amdlibSetDarkData(regionTab, &darkData)
/* DOCUMENT amdlibSetDarkData(regionTab, &darkData)

  DESCRIPTION
    This function replaces the data stored in pixel bias data structure with
    the one placed into the arrays of an amdlibREGION data structure.

  PARAMETERS
    - regionTab : region data structure containing new data values
    - darkData  : structure where pixel bias data has to be stored. It is
                  assumed this structure has been previously allocated.

  RETURN VALUES

  EXAMPLES
    > _amdlibLoadRawData("dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > p=amdlibGetDarkData(&darkData);
    > ...
    > amdlibSetDarkData(p, &darkData);
    > _amdlibReleaseDarkData, &darkData;

  SEE ALSO
    amdlibGetDarkData
*/
{
    /* Copy region data into the raw data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyRegionData(&regionTab, darkData->region, 
                                    darkData->nbRegions, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy data region", errMsg, 3;
    }

    return amdlibSUCCESS;
}

/************************************************************************/

func _amdlibCreateDarkData(&rawData, &darkData, value, errMsg)
/* DOCUMENT _amdlibCreateDarkData(&rawData, &darkData, value, errMsg)

  DESCRIPTION
    This function generates a pixel bias map matching with the data area of
    the raw data and containing the specified value for all pixels.

  PARAMETERS
    - rawData   : structure where raw data is stored
    - darkData  : structure where dark data has to be stored. 
    - value     : value for all pixel of the map
    - errMsg    : if function fails, error message describing failure
    
  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg)
    > _amdlibCreateDarkData(&rawData, darkData, 0, errMsg)
    > _amdlibReleaseRawData, &rawData;
    > p=amdlibGetDarkData(&darkData);
    > pli, (*p(4).data)(, , 1)
    > _amdlibReleaseDarkData, &darkData;

  SEE ALSO
    amdlibSetDarkData
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    darkData = amdlibDARK_DATA();
    status = __amdlibSetDarkData(rawData, &darkData, value, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func amdlibSetFlatFieldMap(value)
/* DOCUMENT amdlibSetFlatFieldMap(value)

  DESCRIPTION
    Set all values to the flat field map to the given value.

  PARAMETERS
    - value : value for all pixel of the map; valid range [0.0-1.0] 

  EXAMPLES
    > amdlibSetFlatFieldMap(0.5);

  SEE ALSO
    _amdlibLoadFlatFieldMap 
*/
{
    status = __amdlibSetFlatFieldMap(value);

    return status;
}

/************************************************************************/

func _amdlibLoadFlatFieldMap(filename, &errMsg)
/* DOCUMENT _amdlibLoadFlatFieldMap(filename, &errMsg)

  DESCRIPTION
    The flat field map contains for each pixel a factor which gives the gain
    of that pixel according to the average gain. The flat field map can be
    loaded from fits files with _amdlibLoadFlatFieldMap().

  PARAMETERS
    - filename : name of file to load
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadFlatFieldMap("FlatFieldMap.fits", errMsg);

  SEE ALSO
    amdlibSetFlatFieldMap, _amdlibLoadBadPixelMap, 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);

    status = __amdlibLoadFlatFieldMap(filename, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibGetWaveDataFromRawData(rawData, &waveData, &errMsg)
/* DOCUMENT _amdlibGetWaveDataFromRawData(rawData, &waveData, &errMsg)

  DESCRIPTION
    Retrieve wave data structure from the specified raw data.

  PARAMETERS
    - rawData  : structure where raw data is stored
    - waveData : structure where wave data will be stored
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > _amdlibGetWaveDataFromRawData(&rawData, waveData, errMsg);
    > _amdlibReleaseRawData, &rawData;

  SEE ALSO
    amdlibGetWaveDataFromP2vm
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    waveData = amdlibWAVEDATA();

    status = __amdlibGetWaveDataFromRawData(rawData, &waveData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func amdlibGetWaveDataFromP2vm(p2vm, &waveData, &errMsg)
/* DOCUMENT amdlibGetWaveDataFromP2vm(p2vm, &waveData, &errMsg)

  DESCRIPTION
    Retrieve wave data structure from the specified p2vm.

  PARAMETERS
    - p2vm     : structure where p2vm is stored
    - waveData : structure where wave data will be stored
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg)
    > amdlibGetWaveDataFromP2vm(&p2vm, waveData, errMsg);
    > _amdlibReleaseP2vm, &p2vm;

  SEE ALSO
    _amdlibGetWaveDataFromRawData
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    waveData = amdlibWAVEDATA();

    status = __amdlibGetWaveDataFromP2vm(p2vm, &waveData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func amdlibGetOiArrayFromRawData(rawData, &oiArray, &errMsg)
/* DOCUMENT amdlibGetOiArrayFromRawData(rawData, &oiArray, &errMsg)

  DESCRIPTION
    Retrieve OI array structure from the specified raw data.

  PARAMETERS
    - rawData : structure where raw data is stored
    - oiArray : structure where OI array information will be stored
    - errMsg  : if function fails, error message describing failure
  
  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which  OI array information is stored has been dynamically
    allocated and must be released using _amdlibReleaseOiArray() when not
    longer used or before futur use, otherwise this will cause unpredictable
    behaviour.

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > amdlibGetOiArrayFromRawData(&rawData, oiArray, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibReleaseOiArray, &oiArray;

  SEE ALSO
    _amdlibReleaseOiArray
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    oiArray = amdlibOI_ARRAY();

    status = __amdlibGetOiArrayFromRawData(rawData, &oiArray, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibAllocateOiArray(nbElements)
/* DOCUMENT _amdlibAllocateOiArray(&OiArray)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - OiArray : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateOiArray(10);

  SEE ALSO
    _amdlibLoadOiArray, amdlibSetOiArray
*/
{
    /* First allocate the error message */
    errMsg = array(char, amdlibERROR_LENGTH);
    /* Create  instance of OI_ARRAY data */
    OiArray = amdlibOI_ARRAY();
    
    status = __amdlibAllocateOiArray(&OiArray, nbElements, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not allocate OI ARRAY", errMsg, 3;
    }
    return OiArray;
}


/************************************************************************/

func _amdlibReleaseOiArray(&oiArray)
/* DOCUMENT _amdlibReleaseOiArray(&oiArray)

  DESCRIPTION
    Release memory allocated by amdlibGetOiArrayFromRawData, to store OI array
    information.

  PARAMETERS
    - oiArray : structure where OI array information is stored
  
  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > amdlibGetOiArrayFromRawData(&rawData, oiArray, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibReleaseOiArray, &oiArray;

  SEE ALSO
    amdlibGetOiArrayFromRawData
*/
{
    status = __amdlibReleaseOiArray(oiArray) ;
} 

/************************************************************************/

func _amdlibGetOiArray(&oiArray)
/* DOCUMENT _amdlibGetOiArray(&oiArray)

  DESCRIPTION
    This function gets the data stored in OI ARRAY data structure and place them
    into a new OI ARRAY data structure which is returned by the function.  This
    new OI ARRAY data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - oiArray : amdlibOI_ARRAY structure with oiArray information stored in
    
  RETURN VALUES
    A amdlibOI_ARRAY structure with oi array information stored in

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > amdlibGetOiArrayFromRawData(&rawData, oiArray, errMsg);
    > oi = _amdlibGetOiArray(&oiArray);
    > _amdlibReleaseOiArray(&oiArray)

  SEE ALSO
    amdlibGetOiArrayFromRawData and _amdlibReleaseOiArray
*/
{
    /* Create new instance of OI_ARRAY data */
    newOiArray = amdlibOI_ARRAY();

    /* Allocate arrays for station descriptions */
    newOiArray.element = &array(amdlibOI_ARRAY_ELEMENT, oiArray->nbStations);

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyOiArray(oiArray, &newOiArray, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy OI array", errMsg, 3;
    }

    return newOiArray;
}

/************************************************************************/

func amdlibGetOiTargetFromRawData(rawData, &oiTarget)
/* DOCUMENT amdlibGetOiTargetFromRawData(rawData, &oiTarget)

  DESCRIPTION
    Retrieve OI target structure from the specified raw data.

  PARAMETERS
    - rawData  : structure where raw data is stored
    - oiTarget : structure where OI array information will be stored

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which  OI array information is stored has been dynamically
    allocated and must be released using _amdlibReleaseOiArray() when not
    longer used or before futur use, otherwise this will cause unpredictable
    behaviour.

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > amdlibGetOiTargetFromRawData(&rawData, oiTarget);
    > targetName = string(&oiTarget.element(1)->targetName)
    > _amdlibReleaseRawData, &rawData;
    > _amdlibReleaseOiTarget, &oiTarget;

  SEE ALSO
    _amdlibReleaseOiTarget
*/
{
    oiTarget = amdlibOI_TARGET();
    oiTarget.nbTargets = 1;
    oiTarget.element = &array(amdlibOI_TARGET_ELEMENT, 1);
    status = __amdlibGetOiTargetFromRawData(rawData, &oiTarget);

    return status;
}

/************************************************************************/

func _amdlibAllocateOiTarget(nbElements)
/* DOCUMENT _amdlibAllocateOiTarget(&OiTarget)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - OiTarget : structure where data will be stored

  RETURN VALUES
    A amdlibOI_TARGET structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateOiTarget(10);

  SEE ALSO
    _amdlibLoadOiTarget, amdlibSetOiTarget
*/
{
    OiTarget = amdlibOI_TARGET();
    status = __amdlibAllocateOiTarget(&OiTarget, nbElements);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate OI TARGET", , 3;
    }
    return OiTarget;
}



/************************************************************************/

func _amdlibReleaseOiTarget(&oiTarget)
/* DOCUMENT _amdlibReleaseOiTarget(&oiTarget)

  DESCRIPTION
    Release memory allocated by amdlibGetOiTargetFromRawData, to store OI
    target information.

  PARAMETERS
    - oiTarget : structure where OI target information is stored
  
  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > amdlibGetOiTargetFromRawData(&rawData, oiTarget, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibReleaseOiTarget, &oiTarget;

  SEE ALSO
    amdlibGetOiTargetFromRawData
*/
{
    status = __amdlibReleaseOiTarget(oiTarget);
}

/************************************************************************/

func _amdlibGetOiTarget(&oiTarget)
/* DOCUMENT _amdlibGetOiTarget(&oiTarget)

  DESCRIPTION
    This function gets the data stored in OI TARGET data structure and place 
    them into a new OI TARGET data structure which is returned by the function.
    This new OI TARGET data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - oiTarget : amdlibOI_TARGET structure with oiTarget information stored in
    
  RETURN VALUES
    A amdlibOI_TARGET structure with oi target information stored in

  EXAMPLES
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > amdlibGetOiTargetFromRawData(&rawData, oiTarget, errMsg);
    > oi = _amdlibGetOiTarget(&oiTarget);
    > _amdlibReleaseOiTarget(&oiTarget)

  SEE ALSO
    amdlibGetOiTargetFromRawData and _amdlibReleaseOiTarget
*/
{
    /* Create new instance of OI_TARGET data */
    newOiTarget = amdlibOI_TARGET();

    /* Allocate arrays for station descriptions */
    newOiTarget.element = &array(amdlibOI_TARGET_ELEMENT, oiTarget->nbTargets);

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyOiTarget(oiTarget, &newOiTarget, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy OI target", errMsg, 3;
    }

    return newOiTarget;
}

/************************************************************************/

func _amdlibRawData2ScienceData(rawData, waveData, &scienceData, sumData, &errMsg)
/* DOCUMENT _amdlibRawData2ScienceData(rawData, waveData, &scienceData, sumData, &errMsg)

  DESCRIPTION
    This function extracts from raw data the science data taking into account
    the offsets between photometric channels and interferometric channel. The
    spectral channels not having entries in interferometrie channel and
    photometric channels are removed. And the photometries for the remaining
    channels is computed by interpolating value according to these offsets.
    The spectral dispersion table is also updated. Only the spectral channels
    which are used are stored in the table. All the tables containing
    interferometric, photometric and spectral dispersion are aligned; i.e.
    they have the same number of entries, and a same entry (same index) in the
    tables corresponds to the same spectral channel.

  PARAMETERS
    - rawData     : structure where raw data is stored
    - waveData    : structure where wave data is stored
    - scienceData : structure where science data will be stored
    - sumData     : resulting science data is the average of all frames
                    contained in raw data
    - errMsg      : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which science data is stored has been dynamically allocated
    and must be released using _amdlibReleaseScienceData() when not longer used
    or before futur use, otherwise this will cause unpredictable behaviour.

  EXAMPLES
    > _amdlibLoadRawData("dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);    
    > _amdlibGetWaveDataFromRawData(&rawData, waveData, errMsg);
    > _amdlibLoadBadPixelMap("BadPixelMap.fits", errMsg);
    > _amdlibLoadFlatFieldMap("FlatFieldMap.fits", errMsg);
    > _amdlibRawData2ScienceData(&rawData, &waveData, scienceData,
    amdlibTRUE, errMsg);
    > _amdlibReleaseRawData, &rawData;

  SEE ALSO
    _amdlibGetScienceData, _amdlibReleaseScienceData
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    scienceData = amdlibSCIENCE_DATA();

    status = __amdlibRawData2ScienceData(rawData, waveData, &scienceData,
                                         sumData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibGetScienceData(&scienceData)
/* DOCUMENT _amdlibGetScienceData(&scienceData)

  DESCRIPTION
    This function gets the data stored in science data structure and place
    them into a new science data structure which is returned by the function.
    This new science data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - scienceData : structure where science data is stored 

  RETURN VALUES
    A amdlibSCIENCE_DATA structure with science data stored in

  EXAMPLES
    > _amdlibLoadRawData("dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData;
    > _amdlibLoadRawData("file.fits", rawData, errMsg)
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);
    > _amdlibGetWaveDataFromRawData(&rawData, waveData,errMsg);
    > _amdlibRawData2ScienceData(&rawData, &waveData, scienceData,
    >                           amdlibFALSE, errMsg);
    > _amdlibReleaseRawData, &rawData
    > s=_amdlibGetScienceData(&scienceData);
    > pli, *(*s.frame)(1).intf
    > plg, *(*s.frame)(1).photo1
    > _amdlibReleaseScienceData, &scienceData;

  SEE ALSO
    amdlibSetScienceData, _amdlibReleaseScienceData
*/
{
    /* Create new instance of science data */
    newScienceData = amdlibSCIENCE_DATA(); 
    /* Allocate arrays for storing science data */
    newScienceData.timeTag = &array(double, scienceData->nbFrames);
    newScienceData.channelNo = &array(int, scienceData->nbChannels);

    newScienceData.frame = &array(amdlibFRAME_SCIENCE_DATA, 
                                  scienceData->nbFrames);
    for (frame = 1; frame <= scienceData->nbFrames; frame++)
    {
        (*newScienceData.frame)(frame).intf = 
            &array(double, scienceData->col(4).nbPixels, 
                   scienceData->nbChannels);
         (*newScienceData.frame)(frame).sigma2Intf = 
            &array(double, scienceData->col(4).nbPixels, 
                   scienceData->nbChannels);
        (*newScienceData.frame)(frame).photo1 = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).photo2 = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).photo3 = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).photoI = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).sigma2Photo1 = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).sigma2Photo2 = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).sigma2Photo3 = 
            &array(double, scienceData->nbChannels);
        (*newScienceData.frame)(frame).sigma2PhotoI = 
            &array(double, scienceData->nbChannels);
    }

    newScienceData.badPixels = &array(char, scienceData->col(4).nbPixels, scienceData->nbChannels);

    /* Copy science data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyScienceData(scienceData, &newScienceData, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy science data", errMsg, 3;
    }

    return newScienceData;
}

/************************************************************************/

func _amdlibReleaseScienceData(&scienceData)
/* DOCUMENT _amdlibReleaseScienceData(&scienceData)

  DESCRIPTION
    Release memory allocated by _amdlibRawData2ScienceData to store science
    data.

  PARAMETERS
    - scienceData : structure to be released
  
  EXAMPLES
    ...
    > _amdlibRawData2ScienceData(&rawData, &waveData, scienceData,
    ...
    > _amdlibReleaseScienceData, &scienceData;

  SEE ALSO
    _amdlibLoadRawData, _amdlibLoadRawDataHdr
*/
{
    status = __amdlibReleaseScienceData(scienceData) 
} 

/************************************************************************/

func amdlibSetScienceData(src, &dst, &errMsg)
/* DOCUMENT amdlibSetScienceData(src, &dst, &errMsg)

  DESCRIPTION
    This function copies the science data structure pointed to by src to the
    structure pointed to by dst.

  PARAMETERS
    - src    : structure where science data to be copied is stored  
    - dst    : structure where science data will be copied
    - errMsg : if function fails, error message describing failure

  EXAMPLES
    > _amdlibLoadRawData(darkFile, rawData, &errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData
    > _amdlibLoadRawData("file.fits", rawData, errMsg)
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);
    > _amdlibGetWaveDataFromRawData(&rawData, waveData,errMsg);
    > _amdlibRawData2ScienceData(&rawData, &waveData, scienceData,
    amdlibFALSE, errMsg);
    > _amdlibReleaseRawData, &rawData
    > s=_amdlibGetScienceData(&scienceData);
    > amdlibSetScienceData(s, &scienceData,errMsg);
    > _amdlibReleaseScienceData, &scienceData;

  SEE ALSO
    _amdlibGetScienceData
*/
{
    /* Copy science data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyScienceData(&src, dst, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy science data", errMsg, 3;
    }

    return amdlibSUCCESS;
}

/************************************************************************/

func _amdlibLoadP2vm(filename, &p2vm, &errMsg)
/* DOCUMENT _amdlibLoadP2vm(filename, &p2vm, &errMsg)

  DESCRIPTION
    The P2VM contains matrix to convert photometry to visibility.

  PARAMETERS
    - filename : name of file to load
    - p2vm     : structure where read P2VM will be stored
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which p2vm is stored has been dynamically allocated
    and must be released using _amdlibReleaseP2vm() when not longer used
    or before futur use, otherwise this will cause unpredictable behaviour.

  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg);
    > _amdlibReleaseP2vm, &p2vm;

  SEE ALSO
    _amdlibGetP2vm, _amdlibReleaseP2vm
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    p2vm = amdlibP2VM_MATRIX();
    status = __amdlibLoadP2VM(filename, &p2vm, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibReleaseP2vm(&p2vm)
/* DOCUMENT _amdlibReleaseP2vm(&p2vm)

  DESCRIPTION
    Release memory allocated by _amdlibLoadP2vm or _amdlibComputeP2vm to store
    P2VM. 

  PARAMETERS
    - p2vm : structure to be released
  
  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg);
    ...
    > _amdlibReleaseP2vm, &p2vm;

  SEE ALSO
    _amdlibLoadP2vm
*/
{
    status = __amdlibReleaseP2VM(p2vm) 
} 

/************************************************************************/

func _amdlibGetP2vm(&p2vm)
/* DOCUMENT _amdlibGetP2vm(&p2vm)

  DESCRIPTION
    This function gets the data stored in P2VM data structure and place them
    into a new P2VM data structure which is returned by the function. This
    new P2VM data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - p2vm : structure where p2vm is stored

  RETURN VALUES
    A amdlibP2VM_MATRIX structure with P2VM data stored in

  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg)
    > p=_amdlibGetP2vm(&p2vm);
    > plg, *p.wlen, color="blue"
    > plg, (*p.sumVk)(, 1), color="red"
    > pli, (*p.vk)(1,,)
    > pli, (*p.badPixels)
    > _amdlibReleaseP2vm, &p2vm;

  SEE ALSO
    _amdlibLoadP2vm
*/
{
    /* Create new instance of science data */
    newP2vm = amdlibP2VM_MATRIX();

    if (p2vm->type == 1)
    {
        nbTel = 2;
        nbBase = 1;
    }
    else
    {
        nbTel = 3;
        nbBase = 3;
    }

    /* Allocate arrays for storing P2VM data */
    newP2vm.wlen = &array(double, p2vm->nbChannels);
    newP2vm.matrix = &array(double, 2 * nbBase, p2vm->nx, p2vm->nbChannels);
    newP2vm.vk = &array(double, p2vm->nx, p2vm->nbChannels, nbTel);
    newP2vm.sumVk = &array(double, p2vm->nbChannels, nbBase);
    newP2vm.photometry = &array(double, p2vm->nbChannels, 3, 2*nbBase+1);
    newP2vm.badPixels = &array(char, p2vm->nx, p2vm->nbChannels);
    newP2vm.flatField = &array(double, p2vm->nx, p2vm->nbChannels);
    newP2vm.flag = &array(char, p2vm->nbChannels);
    newP2vm.phase = &array(double, p2vm->nbChannels, nbBase);

    /* Copy science data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyP2VM(p2vm, &newP2vm, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy P2VM", errMsg, 3;
    }

    return newP2vm;
}

/************************************************************************/

func _amdlibAddToP2vmData(rawData, waveData, p2vmData, &errMsg)
/* DOCUMENT _amdlibAddToP2vmData(rawData, waveData, p2vmData, &errMsg)

  DESCRIPTION
    The _amdlibAddToP2vmData() function extracts the science data from the raw
    data, and stores them into the P2VM data structure.

  PARAMETERS
    - rawData  : structure where raw data is stored
    - waveData : structure where wave data is stored
    - p2vmData : structure where P2VM data will be stored
    - errMsg   : if function fails, error message describing failure
    
  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which P2VM data is stored has been dynamically allocated and
    must be released using _amdlibReleaseP2vmData() when not longer used
    or before futur use, otherwise this will cause unpredictable behaviour.
   
  EXAMPLES
    > p2vmData=amdlibP2VM_INPUT_DATA();
    > _amdlibLoadRawData("file.fits", rawData, errMsg);
    > _amdlibGetWaveDataFromRawData(&rawData, waveData,errMsg);
    > _amdlibAddToP2vmData(&rawData, &waveData, &p2vmData, errMsg);

  SEE ALSO
    _amdlibComputeP2vm, _amdlibReleaseP2vmData
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);

    status = __amdlibAddToP2vmData(rawData, waveData, p2vmData, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibReleaseP2vmData(&p2vmData)
/* DOCUMENT _amdlibReleaseP2vmData(&p2vmData)

  DESCRIPTION
    Release memory allocated by _amdlibLoadP2vm to store P2VM. 

  PARAMETERS
    - p2vmData : structure to be released
  
  EXAMPLES
    > _amdlibAddToP2vmData(&rawData, &waveData, &p2vmData, errMsg);
    ...
    > _amdlibReleaseP2vmData, &p2vmData;

  SEE ALSO
    _amdlibAddToP2vmData
*/
{
    status = __amdlibReleaseP2vmData(p2vmData) 
} 

/************************************************************************/

func _amdlibComputeP2vm(&p2vmData, p2vmType, waveData, &p2vm, &errMsg)
/* DOCUMENT _amdlibComputeP2vm(&p2vmData, p2vmType, waveData, &p2vm, &errMsg)

  DESCRIPTION
    The _amdlibComputeP2vm() compute the P2VM.

  PARAMETERS
    - p2vmData : structure where P2VM data is stored
    - p2vmType : P2VM type; could be either amdlibP2VM_2T or amdlibP2VM_3T
    - waveData : structure where wave data is stored
    - p2vm     : structure where read P2VM will be stored 
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS
    Structure in which p2vm is stored has been dynamically allocated
    and must be released using _amdlibReleaseP2vm() when not longer used
    or before futur use, otherwise this will cause unpredictable behaviour.

  EXAMPLES
    > p2vmData=amdlibP2VM_INPUT_DATA();
    > _amdlibLoadRawData("Dark.fits", rawData, errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData(&rawData);
    > _amdlibLoadBadPixelMap("BadPixelMap.fits", errMsg);
    > _amdlibLoadFlatFieldMap("FlatFieldMap.fits", errMsg);
    > _amdlibLoadRawData("AMBER_2TSTD_CAL115_0001.fits", rawData, errMsg);
    > _amdlibGetWaveDataFromRawData(&rawData, waveData, errMsg);
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);
    > _amdlibAddToP2vmData(&rawData, &waveData, &p2vmData, errMsg);
    > _amdlibReleaseRawData(&rawData);
    > _amdlibLoadRawData("AMBER_2TSTD_CAL115_0002.fits", rawData, errMsg);
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);
    > _amdlibAddToP2vmData(&rawData, &waveData, &p2vmData, errMsg);
    > _amdlibReleaseRawData(&rawData);
    ...
    > _amdlibReleaseDarkData(&darkData);
    > _amdlibComputeP2vm(&p2vmData, amdlibP2VM_2T, &waveData, p2vm, errMsg);
    > _amdlibReleaseP2vmData(&p2vmData); 
    > _amdlibSaveP2vm("newP2vm.fits", &p2vm, amdlibP2VM_STD_ACC, errMsg);
    > _amdlibReleaseP2vm(&p2vm); 

  SEE ALSO
    _amdlibLoadP2vm, _amdlibSaveP2vm 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    p2vm = amdlibP2VM_MATRIX();
    status = __amdlibComputeP2VM(p2vmData, p2vmType, waveData, &p2vm, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibSaveP2vm(filename, &p2vm, accuracy, &errMsg)
/* DOCUMENT _amdlibSaveP2vm(filename, &p2vm, accuracy, &errMsg)

  DESCRIPTION
    The _amdlibSaveP2vm() function saves the P2VM.

  PARAMETERS
    - filename : name of file in which P2VM will be saved
    - p2vm     : structure where P2VM is stored
    - accuracy : P2VM accuracy; could be amdlibP2VM_STD_ACC or
                 amdlibP2VM_HIGH_ACC
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg);
    > _amdlibSaveP2vm("newP2vm.fits", &p2vm, p2vm.accuracy, errMsg)
    > _amdlibReleaseP2vm, &p2vm;

  SEE ALSO
    _amdlibLoadP2vm, _amdlibComputeP2vm 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibSaveP2VM(filename, p2vm, accuracy, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibGetSpectrumFromP2vm(p2vm, wlen, &spectrum, &errMsg)
/* DOCUMENT _amdlibGetSpectrumFromP2vm(p2vm, wlen, &spectrum, &errMsg)

  DESCRIPTION
    Retrieve spectrum from the specified P2VM.

  PARAMETERS
    - p2vm     : structure where p2vm is stored
    - wlen     : OPTINAL, array giving a subset of wavelengths covered by
                 P2VM 
    - spectrum : structure where specrtum data will be stored
    - errMsg   : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg)
    > _amdlibGetSpectrumFromP2vm(&p2vm, , spectrum, errMsg);
    > s = _amdlibGetSpectrum(&spectrum);
    > plg, *s.spec(1);
    > _amdlibReleaseP2vm, &p2vm;
    > _amdlibReleaseSpectrum, &spectrum;

  SEE ALSO
    _amdlibReleaseSpectrum
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    spectrum = amdlibSPECTRUM();

    if (is_void(wlen))
    {
        wavePt = pointer(nil);
    }
    else
    {
        wave = amdlibWAVELENGTH();
        wave.nbWlen = numberof(wlen);
        wave.wlen = &array(double, wave.nbWlen);
        *wave.wlen = wlen;
        wavePt = &wave;
    }

    status = __amdlibGetSpectrumFromP2VM(p2vm, wavePt, &spectrum, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibAllocateSpectrum(nbTels, nbWlen)
/* DOCUMENT _amdlibAllocateSpectrum(&Spectrum)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Spectrum : structure where data will be stored

  RETURN VALUES
    A amdlibSPECTRUM structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateSpectrum(3, 10);

  SEE ALSO
    _amdlibLoadSpectrum, amdlibSetSpectrum
*/
{
    Spectrum = amdlibSPECTRUM();
    status = __amdlibAllocateSpectrum(&Spectrum, nbTels, nbWlen);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate SPECTRUM", , 3;
    }
    return Spectrum;
}

/************************************************************************/

func _amdlibReleaseSpectrum(&spectrum)
/* DOCUMENT _amdlibReleaseSpectrum(&spectrum)

  DESCRIPTION
    Release memory allocated by _amdlibGetSpectrumFromP2vm to store spectrum. 

  PARAMETERS
    - spectrum : structure to be released
  
  EXAMPLES
    > _amdlibGetSpectrumFromP2vm(&p2vm, pointer(nil), spectrum, errMsg);
    > ...
    > _amdlibReleaseSpectrum, &spectrum;

  SEE ALSO
    _amdlibLoadP2vm
*/
{
    status = __amdlibReleaseSpectrum(spectrum);
} 

/************************************************************************/

func _amdlibGetSpectrum(&spectrum)
/* DOCUMENT _amdlibGetSpectrum(&spectrum)

  DESCRIPTION
    This function gets the data stored in spectrum data structure and place them
    into a new spectrum data structure which is returned by the function. This
    new spectrum data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - spectrum : structure where spectrum is stored

  RETURN VALUES
    A amdlibSPECTRUM structure with spectrum data stored in

  EXAMPLES
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg)
    > p = _amdlibGetP2vm(&p2vm);
    > _amdlibGetSpectrumFromP2vm(&p2vm,(*p.wlen)(100:150), spectrum, errMsg);
    > _amdlibReleaseP2vm, &p2vm;
    > s = _amdlibGetSpectrum(&spectrum);
    > plg, *s.spec(1);
    > _amdlibReleaseSpectrum, &spectrum;

  SEE ALSO
    _amdlibGetSpectrumFromP2vm
*/
{
    /* Create new instance of science data */
    newSpectrum = amdlibSPECTRUM();

    /* Allocate arrays for storing P2VM data */
    for (tel = 1; tel <= spectrum->nbTels; tel++)
    {
        newSpectrum.spec(tel) = &array(double, spectrum->nbWlen);
        newSpectrum.specErr(tel) = &array(double, spectrum->nbWlen);
    }
    newSpectrum.specI = &array(double, spectrum->nbWlen);
    newSpectrum.specErrI = &array(double, spectrum->nbWlen);

    /* Copy science data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopySpectrum(spectrum, &newSpectrum, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy spectrum", errMsg, 3;
    }

    return newSpectrum;
}

/************************************************************************/

func _amdlibLoadSelection(filename, &selection, &errMsg)
/* DOCUMENT _amdlibLoadSelection(filename, &selection, &errMsg)

  DESCRIPTION
    This routine allows to load a file containing flags corresponding to the
    good frames after a frame selection.

  PARAMETERS
    - filename  : name of file to load
    - selection : structure where read data will be stored
    - errMsg    : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibLoadSelection("file.fits", selection, errMsg);
    ...
    > _amdlibReleaseSelection(&selection);

  SEE ALSO
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    insCfg = amdlibINS_CFG();
    selection = amdlibSELECTION();

    status = __amdlibReadSelectionFile(filename, &insCfg, &selection, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibWriteSelectionFile(filename, &selection, &errMsg)
/* DOCUMENT _amdlibWriteSelectionFile(filename, &selection, &errMsg)

  DESCRIPTION
    This routine allows to load a file containing flags corresponding to the
    good frames after a frame selection.

  PARAMETERS
    - filename  : name of file to load
    - selection : structure where read data will be stored
    - errMsg    : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > _amdlibWriteSelectionFile("file.fits", selection, errMsg);
    ...
    > _amdlibReleaseSelection(&selection);

  SEE ALSO
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);

    status = __amdlibWriteSelectionFile(filename, selection, &errMsg);
    errMsg = string(&errMsg);

    return status;
}

/************************************************************************/

func _amdlibReleaseSelection(&selection)
/* DOCUMENT _amdlibReleaseSelection(&selection)

  DESCRIPTION
    Release memory allocated by _amdlibLoadSelection to store selection data.

  PARAMETERS
    - selection : structure to be released
  
  EXAMPLES
    > _amdlibLoadSelection("file.fits", selection, errMsg);
    ...
    > _amdlibReleaseSelection, &selection;

  SEE ALSO
    _amdlibLoadRawData, _amdlibLoadRawDataHdr
*/
{
    status = __amdlibReleaseSelection(selection) 
} 

/************************************************************************/

func amdlibComputeVisibilities(&data, &p2vm, &waveData, band, &visOptions, &photometry, &vis, &vis2, &vis3, &wave, &opd, &errMsg)
/* DOCUMENT amdlibComputeVisibilities(&data, &p2vm, &waveData, band, &visOptions, &photometry, &vis, &vis2, &vis3, &wave, &opd, &errMsg)

  DESCRIPTION
    The amdlibComputeVisibilities() compute the raw visibilities and pistons.

  PARAMETERS
    - data       : structure where science data is stored
    - p2vm       : structure where P2VM is stored
    - waveData   : structure where wavelength information is stored
    - band       : band the spectral band to select for computation; could
                   amdlibJ_BAND, amdlibH_BAND or amdlibK_BAND
    - visOptions : amdlibCPT_VIS_OPTIONS
    - photometry : structure in which will be stored the (binned) photometry
                   results.
    - vis        : structure in which will be stored the (binned) raw
                   visibilities.
    - vis2       : structure in which will be stored the (binned) squared
                   visibilities.
    - vis3       : structure in which will be stored the (binned) triple
                   products.
    - wave       : structure in which will be stored the wavelengths
                   corresponding to the other output structures.
    - opd        : structure in which will be stored the (binned) piston
                   measurements.
    - errMsg     : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS 
    Structure in which results are stored have been dynamically allocated and
    must be released using _amdlibReleasePhotometry(), _amdlibReleaseVis(),
    _amdlibReleaseVis2(), _amdlibReleaseVis3(), _amdlibReleaseWavelength() and
    _amdlibReleasePiston() when not longer used or before futur use, otherwise
    this will cause unpredictable behaviour.

  EXAMPLES
    > _amdlibLoadRawData("dark.fits", rawData, &errMsg);
    > _amdlibGenerateDarkData(rawData, darkData, errMsg);
    > _amdlibReleaseRawData, &rawData
    > _amdlibLoadP2vm("p2vm.fits", p2vm, errMsg)
    > _amdlibLoadRawData("AMBER_2TSTD_OBS193_0004.fits.gz", rawData, &errMsg);
    > _amdlibCalibrateRawData(&darkData, &rawData, &errMsg);
    > _amdlibGetWaveDataFromRawData(&rawData, waveData, errMsg);
    > amdlibGetOiArrayFromRawData(&rawData, oiArray, errMsg);
    > _amdlibRawData2ScienceData(&rawData, &waveData, scienceData,
    amdlibFALSE, errMsg); 
    > amdlibGetOiTargetFromRawData(&scienceData, oiTarget);
    > _amdlibReleaseRawData(&rawData);
    > visOptions = amdlibCPT_VIS_OPTIONS();
    > visOptions.binSize = 999999;
    > visOptions.errorType = amdlibSTATISTICAL_ERROR;
    > visOptions.pistonType = amdlibUNWRAPPED_PHASE;
    > visOptions.noCheckP2vmId = amdlibTRUE;
    > visOptions.frameSelectionType = amdlibNO_FRAME_SEL;
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, 
    vis, vis2, vis3, wave, opd, errMsg);
    > v=_amdlibGetVis(&vis);
    > plg, (*(*v.table)(1, 1).vis)().re
    > plg, (*(*v.table)(1, 1).vis)().im
    > v2=_amdlibGetVis2(&vis2);
    > plg, *(*v2.table)(1, 1).vis2
    > v3=_amdlibGetVis3(&vis3);
    > plg, *(*v3.table)(1, 1).vis3Amplitude
    > p=_amdlibGetPiston(&opd);
    > plg, (*p.pistonOPD)(1, )
    > plg, (*p.sigmaPiston)(1, )
    > _amdlibReleasePhotometry(&phot)
    > _amdlibReleaseVis(&vis)
    > _amdlibReleaseVis2(&vis2)
    > _amdlibReleaseVis3(&vis3)
    > _amdlibReleaseWavelength(&wave)
    > _amdlibReleasePiston(&opd)

  SEE ALSO
    _amdlibReleasePhotometry, _amdlibReleaseVis, _amdlibReleaseVis2,
    _amdlibReleaseVis3, _amdlibReleaseWavelength, _amdlibReleasePiston

*/
{
    errMsg = array(char, amdlibERROR_LENGTH);
    photometry = amdlibPHOTOMETRY();
    vis = amdlibVIS();
    vis2 = amdlibVIS2();
    vis3 = amdlibVIS3();
    wave = amdlibWAVELENGTH();
    opd = amdlibPISTON();

    status = __amdlibComputeVisibilities(data, p2vm, waveData, band,
                                         visOptions, &photometry, &vis, &vis2,
                                         &vis3, &wave, &opd, &errMsg);
    errMsg = string(&errMsg);
    return status;
}


/************************************************************************/

func _amdlibAllocatePhotometry(nbBases, nbFrames, nbWlen)
/* DOCUMENT _amdlibAllocatePhot(&Phot)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Phot : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocatePhotometry(3, 10, 10);

  SEE ALSO
    _amdlibLoadPhot, amdlibSetPhot
*/
{
    Phot = amdlibPHOTOMETRY();
    status = __amdlibAllocatePhotometry(&Phot, nbFrames, nbBases, nbWlen);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate PHOTOMETRY table", , 3;
    }
    return Phot;
}


/************************************************************************/

func _amdlibReleasePhotometry(&phot)
/* DOCUMENT _amdlibReleasePhotometry(&phot)

  DESCRIPTION
    Release memory allocated by amdlibComputeVisibilities or amdlibLoadOiFile,
    to store photometry.

  PARAMETERS
    - phot : structure where photometry is stored
  
  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > _amdlibReleasePhotometry, &phot;

  SEE ALSO
    amdlibComputeVisibilities, amdlibLoadOiFile
*/
{
    status = __amdlibReleasePhotometry(phot) ;
}

/************************************************************************/

func _amdlibAllocateVis(nbBases, nbFrames, nbWlen)
/* DOCUMENT _amdlibAllocateVis(&Vis)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Vis : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateVis(10);

  SEE ALSO
    _amdlibLoadVis, amdlibSetVis
*/
{
    Vis = amdlibVIS();
    status = __amdlibAllocateVis(&Vis, nbFrames, nbBases, nbWlen);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate OI VIS", , 3;
    }
    return Vis;
}



/************************************************************************/

func _amdlibReleaseVis(&vis)
/* DOCUMENT _amdlibReleaseVis(&vis)

  DESCRIPTION
    Release memory allocated by amdlibComputeVisibilities or amdlibLoadOiFile,
    to store visibility.

  PARAMETERS
    - vis : structure where visibility is stored
  
  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > _amdlibReleaseVis, &vis;

  SEE ALSO
    amdlibComputeVisibilities, amdlibLoadOiFile
*/
{
    status = __amdlibReleaseVis(vis);
}

/************************************************************************/

func _amdlibAllocateVis2(nbBases, nbFrames, nbWlen)
/* DOCUMENT _amdlibAllocateVis2(&Vis2)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Vis2 : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateVis2(10);

  SEE ALSO
    _amdlibLoadVis2, amdlibSetVis2
*/
{
    Vis2 = amdlibVIS2();
    status = __amdlibAllocateVis2(&Vis2, nbFrames, nbBases, nbWlen);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate OI VIS2", , 3;
    }
    return Vis2;
}


/************************************************************************/

func _amdlibReleaseVis2(&vis2)
/* DOCUMENT _amdlibReleaseVis2(&vis2)

  DESCRIPTION
    Release memory allocated by amdlibComputeVisibilities or amdlibLoadOiFile,
    to store squared visibility.

  PARAMETERS
    - vis2 : structure where squared visibility is stored
  
  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > _amdlibReleaseVis2, &vis2;

  SEE ALSO
    amdlibComputeVisibilities, amdlibLoadOiFile
*/
{
    status = __amdlibReleaseVis2(vis2) ;
} 


/************************************************************************/

func _amdlibAllocateVis3(nbBases, nbFrames, nbWlen)
/* DOCUMENT _amdlibAllocateVis3(&Vis3)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Vis3 : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateVis3(10);

  SEE ALSO
    _amdlibLoadVis3, amdlibSetVis3
*/
{
    Vis3 = amdlibVIS3();
    status = __amdlibAllocateVis3(&Vis3, nbFrames, nbBases, nbWlen);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate OI VIS3", , 3;
    }
    return Vis3;
}

/************************************************************************/

func _amdlibReleaseVis3(&vis3)
/* DOCUMENT _amdlibReleaseVis3(&vis3)

  DESCRIPTION
    Release memory allocated by amdlibComputeVisibilities or amdlibLoadOiFile,
    to store phase closure.

  PARAMETERS
    - vis3 : structure where phase closure is stored
  
  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > _amdlibReleaseVis3, &vis3;

  SEE ALSO
    amdlibComputeVisibilities, amdlibLoadOiFile
*/
{
    status = __amdlibReleaseVis3(vis3) ;
} 

/************************************************************************/

func _amdlibAllocateWavelength(nbWlen)
/* DOCUMENT _amdlibAllocateWavelength(&Wavelength)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Wavelength : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocateWavelength(10);

  SEE ALSO
    _amdlibLoadWavelength, amdlibSetWavelength
*/
{
    Wavelength = amdlibWAVELENGTH();
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibAllocateWavelength(&Wavelength, nbWlen, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not allocate OI WAVELENGTH", errMsg, 3;
    }
    return Wavelength;
}


/************************************************************************/

func _amdlibReleaseWavelength(&wave)
/* DOCUMENT _amdlibReleaseWavelength(&wave)

  DESCRIPTION
    Release memory allocated by amdlibComputeVisibilities or amdlibLoadOiFile,
    to store wavelengths.

  PARAMETERS
    - wave : structure where wavelengths are stored
  
  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > _amdlibReleaseWavelength, &wave;

  SEE ALSO
    amdlibComputeVisibilities, amdlibLoadOiFile
*/
{
    status = __amdlibReleaseWavelength(wave) ;
} 

/************************************************************************/

func _amdlibAllocatePiston(nbBases, nbFrames)
/* DOCUMENT _amdlibAllocatePiston(&Piston)

  DESCRIPTION
    This function allocates the OI ARRAY amdlib structure.
    
  PARAMETERS
    - Piston : structure where data will be stored

  RETURN VALUES
    A amdlibOI_ARRAY structure with the right dimensions , filled with zeros
    
  EXAMPLES
    > array = _amdlibAllocatePiston(10);

  SEE ALSO
    _amdlibLoadPiston, amdlibSetPiston
*/
{
    Piston = amdlibPISTON();
    status = __amdlibAllocatePiston(&Piston, nbFrames, nbBases);
    if (status == amdlibFAILURE)
    {
        yocoError, "Could not allocate PISTON", , 3;
    }
    return Piston;
}


/************************************************************************/

func _amdlibReleasePiston(&opd)
/* DOCUMENT _amdlibReleasePiston(&opd)

  DESCRIPTION
    Release memory allocated by amdlibComputeVisibilities or amdlibLoadOiFile,
    to store OPD information.

  PARAMETERS
    - opd : structure where OPD information is stored
  
  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData,
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > _amdlibReleasePiston, &opd;

  SEE ALSO
    amdlibComputeVisibilities, amdlibLoadOiFile
*/
{
    status = __amdlibReleasePiston(opd) 
} 

/************************************************************************/

func amdlibWriteOiFile(filename, &insCfg, &oiArray, &oiTarget, &wave, &photometry, &vis, &vis2, &vis3, &opd, &spectrum, &errMsg)
/* DOCUMENT amdlibWriteOiFile(filename, &insCfg, &oiArray, &oiTarget, &wave, &photometry, &vis, &vis2, &vis3, &opd, &spectrum, &errMsg)
    
  DESCRIPTION
    This function writes OI files from the AMBER data contained in the amdlib
    structures

  PARAMETERS
    - filename   : name of file in which OI data will be saved 
    - insCfg     : instrument configuration keyword list 
    - oiArray    : structure where OI array information is stored
    - oiTarget   : structure where OI target information is stored
    - wave       : structure in which is stored the wavelengths corresponding
                   to the other output structures.
    - photometry : structure in which is stored the (binned) photometry
    - vis        : structure in which is stored the (binned) raw visibilities.
    - vis2       : structure in which is stored the (binned) squared
                   visibilities.
    - vis3       : structure in which is stored the (binned) triple products.
    - opd        : structure in which is stored the (binned) piston
                   measurements.
    - spectrum   : structure in which is stored the spectra.
    - errMsg     : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  EXAMPLES
    > amdlibWriteOiFile("oi.fits", &scienceData.insCfg, &oiArray, &oiTarget,
    &wave, &photometry, &vis, &vis2, &vis3, &opd, 
    &spectrum, errMsg);

  SEE ALSO
    _amdlibReadOiFile 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);

    status = __amdlibSaveOiFile(filename, 
                                insCfg,
                                oiArray, 
                                oiTarget, 
                                photometry, 
                                vis, 
                                vis2, 
                                vis3, 
                                wave, 
                                opd, 
                                spectrum,
                                &errMsg);
    errMsg = string(&errMsg);

    return status;
} 

/************************************************************************/

func _amdlibReadOiFile(filename, &insCfg, &oiArray, &oiTarget, &wave, &photometry, &vis, &vis2, &vis3, &opd, &spectrum, &errMsg)
/* DOCUMENT _amdlibReadOiFile(filename, &insCfg, &oiArray, &oiTarget, &wave, &photometry, &vis, &vis2, &vis3, &opd, &spectrum, &errMsg)
  
  DESCRIPTION
    This function reads OI files and stored read informations in the amdlib
    structures
   
  PARAMETERS
    - filename   : name of file in which OI data to read 
    - insCfg     : instrument configuration keyword list 
    - oiArray    : structure where OI array information will be stored
    - oiTarget   : structure where OI target information will be stored
    - wave       : structure in which will be stored the wavelengths
                   corresponding to the other output structures.
    - photometry : structure in which will be stored the (binned) photometry
    - vis        : structure in which will be stored the (binned) raw
                   visibilities.
    - vis2       : structure in which will be stored the (binned) squared
                   visibilities.
    - vis3       : structure in which will be stored the (binned) triple
                   products.
    - opd        : structure in which will be stored the (binned) piston
                   measurements.
    - spectrum   : structure in which will be stored the spectra.
    - errMsg     : if function fails, error message describing failure

  RETURN VALUES
    - amdlibSUCCESS upon successful completion
    - amdlibFAILURE otherwise

  CAUTIONS 
    Structure in which results are stored have been dynamically allocated and
    must be released using _amdlibReleasePhotometry(), _amdlibReleaseVis(),
    _amdlibReleaseVis2(), _amdlibReleaseVis3(), _amdlibReleaseWavelength() and
    _amdlibReleasePiston() when not longer used or before futur use, otherwise
    this will cause unpredictable behaviour.

  EXAMPLES
    > _amdlibReadOiFile("oi.fits", insCfg, oiArray, oiTarget, wave, photometry, vis, vis2, vis3, opd, spectrum, errMsg);

  SEE ALSO
    _amdlibReleasePhotometry, _amdlibReleaseVis, _amdlibReleaseVis2,
    _amdlibReleaseVis3, _amdlibReleaseWavelength, _amdlibReleasePiston,
    amdlibSaveOiFile 
*/
{
    errMsg = array(char, amdlibERROR_LENGTH);

    insCfg = amdlibINS_CFG();
    oiArray = amdlibOI_ARRAY();
    oiTarget = amdlibOI_TARGET();
    wave = amdlibWAVELENGTH();
    photometry = amdlibPHOTOMETRY();
    vis = amdlibVIS();
    vis2 = amdlibVIS2();
    vis3 = amdlibVIS3();
    opd = amdlibPISTON();
    spectrum = amdlibSPECTRUM();
    status = __amdlibLoadOiFile(filename, 
                                &insCfg,
                                &oiArray, 
                                &oiTarget, 
                                &photometry,
                                &vis, 
                                &vis2, 
                                &vis3,
                                &wave,
                                &opd, 
                                &spectrum,
                                &errMsg);
    errMsg = string(&errMsg);

    if (status == amdlibFAILURE)
    {
        /* Release the memory */
        __amdlibReleasePhotometry, &photometry;
        __amdlibReleasePiston, &opd;
        __amdlibReleaseVis2, &vis2;
        if (vis.nbBases == 3)
        {
            __amdlibReleaseVis3, &vis3;
        }
        __amdlibReleaseVis, &vis;
        __amdlibReleaseWavelength, &wave;
        __amdlibReleasePiston,&opd;
        __amdlibReleaseOiArray, &oiArray;
        __amdlibReleaseOiTarget, &oiTarget;
        __amdlibReleaseSpectrum, &spectrum;
    }

    return status;
} 

/************************************************************************/

func _amdlibGetPhot(&phot)
/* DOCUMENT _amdlibGetPhot(&phot)

  DESCRIPTION
    This function gets the data stored in PHOTOMETRY data structure and place
    them into a new PHOTOMETRY data structure which is returned by the
    function.  This new PHOTOMETRY data structure is allocated by this
    procedure providing convenient access to data arrays inside Yorick
    programs.

  PARAMETERS
    - phot : structure where are stored the telescopes fluxes
    
  RETURN VALUES
    A amdlibPHOTOMETRY structure with photometry stored in

  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData, 
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > p = _amdlibGetPhot(&photometry);

  SEE ALSO
    amdlibComputeVisibilities and _amdlibReadOiFile
*/
{
    local newPhot;
    
    /* Create new instance of PHOTOMETRY data */
    newPhot = amdlibPHOTOMETRY();
    
    /* Allocate arrays for Photometries */
    newPhot.table = &array(amdlibPHOTOMETRY_TABLE_ENTRY, phot->nbBases, 
                           phot->nbFrames);

    for (base = 1; base <= phot->nbBases; base++)
    { 
        for (frame = 1; frame <= phot->nbFrames; frame++)
        {
            (*newPhot.table)(base, frame).fluxSumPiPj =
                &array(double, phot->nbWlen);

            (*newPhot.table)(base, frame).sigma2FluxSumPiPj =
                &array(double, phot->nbWlen);

            (*newPhot.table)(base, frame).fluxRatPiPj =
                &array(double, phot->nbWlen);

            (*newPhot.table)(base, frame).sigma2FluxRatPiPj =
                &array(double, phot->nbWlen);

            (*newPhot.table)(base, frame).PiMultPj =
                &array(double, phot->nbWlen);
        }
    }

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyPhotometry(phot, &newPhot, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy photometry", errMsg, 3;
    }

    yocoLogTrace, "Phot get successfully";

    return newPhot;
}

/************************************************************************/

func _amdlibGetVis(&vis)
/* DOCUMENT _amdlibGetVis(&vis)

  DESCRIPTION
    This function gets the data stored in VIS data structure and place them
    into a new VIS data structure which is returned by the function.  This new
    VIS data structure is allocated by this procedure providing convenient
    access to data arrays inside Yorick programs.

  PARAMETERS
    - vis : amdlibVIS structure with visibilities stored in
    
  RETURN VALUES
    A amdlibVIS structure with visibilities stored in

  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData, 
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > v = _amdlibGetVis(&vis);

  SEE ALSO
    amdlibComputeVisibilities and _amdlibReadOiFile    
*/
{
    local newVis;
    /* Create new instance of VIS data */
    newVis = amdlibVIS();

    /* Allocate arrays for visibilities */
    newVis.table = &array(amdlibVIS_TABLE_ENTRY, vis->nbBases, vis->nbFrames);

    for (base = 1; base <= vis->nbBases; base++)
    { 
        for (frame = 1; frame <= vis->nbFrames; frame++)
        {
            (*newVis.table)(base, frame).vis =
                &array(amdlibCOMPLEX, vis->nbWlen);

            (*newVis.table)(base, frame).sigma2Vis =
                &array(amdlibCOMPLEX, vis->nbWlen);

            (*newVis.table)(base, frame).visCovRI =
                &array(double, vis->nbWlen);

            (*newVis.table)(base, frame).diffVisAmp =
                &array(double, vis->nbWlen);

            (*newVis.table)(base, frame).diffVisAmpErr =
                &array(double, vis->nbWlen);

            (*newVis.table)(base, frame).diffVisPhi =
                &array(double, vis->nbWlen);

            (*newVis.table)(base, frame).diffVisPhiErr =
                &array(double, vis->nbWlen);
            
            (*newVis.table)(base, frame).chiSquare =
                &array(double, vis->nbWlen);
            
            (*newVis.table)(base, frame).flag =
                &array(int, vis->nbWlen);

        }
    }

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyVis(vis, &newVis, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy vis", errMsg, 3;
    }

    yocoLogTrace, "vis get successfully";

    return newVis;
}

/************************************************************************/

func _amdlibGetVis2(&vis2)
/* DOCUMENT _amdlibGetVis2(&vis2)

  DESCRIPTION
    This function gets the data stored in VIS2 data structure and place them
    into a new VIS2 data structure which is returned by the function.  This
    new VIS2 data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - vis2 : amdlibVIS2 structure with squared visibilities stored in
    
  RETURN VALUES
    A amdlibVIS2 structure with squared visibilities stored in

  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData, 
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > v = _amdlibGetVis2(&vis2);

  SEE ALSO
    amdlibComputeVisibilities and _amdlibReadOiFile    
*/
{
    /* Create new instance of VIS2 data */
    newVis2 = amdlibVIS2();

    /* Allocate arrays for visibilities */
    newVis2.table = &array(amdlibVIS2_TABLE_ENTRY, 
                           vis2->nbBases, vis2->nbFrames);
    for (base = 1; base <= vis2->nbBases; base++)
    { 
        for (frame = 1; frame <= vis2->nbFrames; frame++)
        {
            (*newVis2.table)(base, frame).vis2 = &array(double, vis2->nbWlen);
            
            (*newVis2.table)(base, frame).vis2Error =
                &array(double, vis2->nbWlen);
            (*newVis2.table)(base, frame).flag = &array(int, vis2->nbWlen);
        }
    }

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyVis2(vis2, &newVis2, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy vis2", errMsg, 3;
    }

    yocoLogTrace, "vis2 get successfully";

    return newVis2;
}

/************************************************************************/

func _amdlibGetVis3(&vis3)
/* DOCUMENT _amdlibGetVis3(&vis3)

  DESCRIPTION
    This function gets the data stored in VIS3 data structure and place them
    into a new VIS3 data structure which is returned by the function.  This
    new VIS3 data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - vis3 : amdlibVIS3 structure with phase closures stored in
    
  RETURN VALUES
    A amdlibVIS3 structure with phase closures stored in

  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData, 
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > v = _amdlibGetVis3(&vis3);

  SEE ALSO
    amdlibComputeVisibilities and _amdlibReadOiFile    
*/
{
    /* Create new instance of VIS3 data */
    newVis3 = amdlibVIS3();

    /* Allocate arrays for visibilities */
    newVis3.table = &array(amdlibVIS3_TABLE_ENTRY, 
                           vis3->nbClosures, vis3->nbFrames);
    for (clos = 1; clos <= vis3->nbClosures; clos++)
    { 
        for (frame = 1; frame <= vis3->nbFrames; frame++)
        { 
            (*newVis3.table)(clos, frame).vis3Amplitude = 
                &array(double, vis3->nbWlen);
            
            (*newVis3.table)(clos, frame).vis3AmplitudeError =
                &array(double, vis3->nbWlen);
            
            (*newVis3.table)(clos, frame).vis3Phi = 
                &array(double, vis3->nbWlen);
            
            (*newVis3.table)(clos, frame).vis3PhiError = 
                &array(double, vis3->nbWlen);
            (*newVis3.table)(clos, frame).flag = 
                &array(int, vis3->nbWlen);
        }
    }

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyVis3(vis3, &newVis3, &errMsg);
    if (status == amdlibFAILURE)
    { 
        errMsg = string(&errMsg);
        yocoError, "Could not copy vis3", errMsg, 3;
    }

    yocoLogTrace, "vis3 get successfully";

    return newVis3;
}

/************************************************************************/

func _amdlibGetWavelength(&wave)
/* DOCUMENT _amdlibGetWavelength(&wave)

  DESCRIPTION
    This function gets the data stored in WAVELENGTH data structure and place 
    them into a new WAVELENGTH data structure which is returned by the function
    This new WAVELENGTH data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - wave : amdlibWAVELENGTH structure with wavelength information stored in
    
  RETURN VALUES
    A amdlibWAVELENGTH structure with wavelength information stored in

  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData, 
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > w = _amdlibGetWavelength(&wave);

  SEE ALSO
    amdlibComputeVisibilities and _amdlibReadOiFile
*/
{
    /* Create new instance of WAVELENGTH data */
    newWlen = amdlibWAVELENGTH();

    /* Allocate arrays for visibilities */
    newWlen.nbWlen = wave->nbWlen;

    newWlen.wlen = &array(double, wave->nbWlen);
    newWlen.bandwidth = &array(double, wave->nbWlen);

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyWavelength(wave, &newWlen, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy wavelength", errMsg, 3;
    }

    yocoLogTrace, "wlen get successfully";

    return newWlen;
}

/************************************************************************/

func _amdlibGetPiston(&opd)
/* DOCUMENT _amdlibGetPiston(&opd)

  DESCRIPTION
    This function gets the data stored in PISTON data structure and place them
    into a new PISTON data structure which is returned by the function.  This
    new PISTON data structure is allocated by this procedure providing
    convenient access to data arrays inside Yorick programs.

  PARAMETERS
    - opd : amdlibPISTON structure with piston information stored in
    
  RETURN VALUES
    A amdlibPISTON structure with OPD stored in

  EXAMPLES
    > amdlibComputeVisibilities(&scienceData, &p2vm, &waveData, 
    amdlibK_BAND, &visOptions, photometry, vis,
    vis2, vis3, wave, opd, errMsg);
    > p = _amdlibGetPiston(&opd);

  SEE ALSO
    amdlibComputeVisibilities and _amdlibReadOiFile    
*/
{
    /* Create new instance of PISTON data */
    newOpd = amdlibPISTON();

    /* Allocate arrays for piston */
    for (base = 1; base <= 3; base++)
    {
        newOpd.pistonOPDArray(base) = 
            &array(double, opd->nbBases, opd->nbFrames);
        newOpd.sigmaPistonArray(base) =
            &array(double, opd->nbBases, opd->nbFrames);
    }
    newOpd.pistonOPD = &array(double, opd->nbBases, opd->nbFrames);
    newOpd.sigmaPiston = &array(double, opd->nbBases, opd->nbFrames);

    /* Copy data structure */
    errMsg = array(char, amdlibERROR_LENGTH);
    status = __amdlibCopyPiston(opd, &newOpd, &errMsg);
    if (status == amdlibFAILURE)
    {
        errMsg = string(&errMsg);
        yocoError, "Could not copy piston", errMsg, 3;
    }

    yocoLogTrace, "piston get successfully";

    return newOpd;
}

func _amdlibGetCVersion(void)
/* DOCUMENT _amdlibGetCVersion()

  DESCRIPTION  
    Returns the value of version associated to the first found amdlib C program.

  RETURN VALUE
    The version of C programs

  EXAMPLE
    > v = _amdlibGetCVersion();

  SEE ALSO
    _amdlibCheckRuntimeVersions
*/
{
    /*  Extract information from first line of some C executables */
    /*  pattern should be 'amdlib, version VVVVVV XXXX'  see amdlibPrintVersion() */
    tmpFilename = "/tmp/tmpAmdlibVersion";
    system,"amdlibComputeOiData -h | head -1 > " + tmpFilename;
    yocoFileReadAscii, tmpFilename, v;
    system, "rm " + tmpFilename;
    return v(3);
}

func _amdlibCheckRuntimeVersions(void)
/* DOCUMENT _amdlibCheckRuntimeVersions()

  DESCRIPTION  
    Warn user if C and Yorick programs use different version.

  RETURN VALUE
    1 if versions are the same, 0 else

  EXAMPLE
    > v = _amdlibCheckRuntimeVersions();

  SEE ALSO
    _amdlibCheckRuntimeVersions
*/
{
  /* Get C version */
  cVersion = strtrim(_amdlibGetCVersion());

  /* Get yorick version */
  yorickVersion=int();
  amdlibVersion, yorickVersion;
  
  /* Compare strings */
  if (cVersion != yorickVersion)
  {
      yocoLogWarning, "Possible version conflict:";
      yocoLogWarning, "C part is '" + cVersion + "'  while yorick one is '" + yorickVersion + "'";
      return 0;
  }
  return 1;
}



/********************* GLOBAL VARIABLES ***********************************/

// Bad pixel and flat field files
amdlibBadPixelFile = [];
amdlibFlatFieldFile = [];

// Set the copyright here !
amdlibNO_COPYRIGHT = 0; 


/********************* CONVERSION FACTORS ******************************/

angstrom2metre = 1e-10;
min2deg = 1./60.;
sec2deg = min2deg / 60.;
mas2deg = sec2deg / 1000;
deg2rad = pi/180.;
rad2deg = 180./pi;
mas2rad = mas2deg * deg2rad;
sec2rad = sec2deg * deg2rad;
rad2mas = rad2deg / mas2deg;
C       = 2.99792453e+08;

unit2kilo = milli2unit = 1e-3;
unit2mega = micro2unit = 1e-6;
unit2giga = nano2unit  = 1e-9;
unit2tera = pico2unit  = 1e-12;

kilo2unit = unit2milli = 1e3;
mega2unit = unit2micro = 1e6;
giga2unit = unit2nano  = 1e9;
tera2unit = unit2pico  = 1e12;

/********************* INITIALISATION *************************************/

// Get working directory and home directory
cwd = get_cwd();
HOME = cd("~");
cd, cwd;

// Set random variables to random ;-)
randomize;

// Default for graphics
pldefault, legends=1, style="amdlib1horiz.gs", marks=0;

// Get current software version
version=int();
amdlibVersion, version;
write, "\n --==< amdlib, version " + version + " >==--";
_amdlibCheckRuntimeVersions;
write, "Type \"amdlib\" if you want some help\n";

// Write acknowlegdement request
write, "\\\\\\|||||||||||||||||||||||||||||||||||||||||||||||||||||||///";
write, "---                                                       ---";
write, "---       PLEASE ACKNOWLEDGE THE USE OF AMBER DATA        ---";
write, "---          REDUCTION SOFTWARE FROM THE JMMC,            ---";
write, "---           FOR EACH USE IN A PUBLICATION.              ---";
write, "---      http://www.jmmc.fr/amberdrs/acknowledgement      ---";
write, "---                                                       ---";
write, "///|||||||||||||||||||||||||||||||||||||||||||||||||||||||\\\\\\\n";

